/*
  Demonstrating how to control RTL-SDR USB dongle using SDRSharp,
  Net Remote plugin, Arduino Uno MCU plus Ky-40 rotary encoder.
  
  SDRSharp Net Remote copyright 2015 Al Brown. Original Arduino
  sample program by Al Brown.
  
  Arduino MCU program, modifications & enhancements by
  Robert C. Mazur, VA3ROM, Feb. 2016 (Arduino IDE 1.6.5 used)
  SDR# v1.1430 (dot Net 4.5 version) used
  
  Requires:
      Net Remote SDR# plugin
      Rotary encoder library for Arduino IDE
      JSON library for Arduino IDE
      
      http://eartoearoak.com/software/sdrsharp-net-remote
      http://www.pjrc.com/teensy/td_libs_Encoder.html
      https://github.com/bblanchon/ArduinoJson
  
  Rotary encoder and shaft push button connects:
      Encoder/KY-40        Arduino
      Pin A/CLK            D3
      Pin B/DT             D2
      Pin C/GND            GND

      KY-40 specific       Arduino
      SW                   D11
      VCC                  Arduino +5 VDC
      
  Uses LCD/keypad shield (SainSmart v1.1 compatible or compatible clones).
  Uses RTL-SDR USB dongle (with R820T or R820T2 tuner).
  Uses SDR# (dotNet 4.5 version) with compatible plugins.
  
  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  GNU General Public License, see <http://www.gnu.org/licenses/>.
*/

/***************************************************************************************************
 *                              Required Program Include Support Files                             *
 ***************************************************************************************************/

#include <ArduinoJson.h>                                 // JSON support library
#include <Encoder.h>                                     // rotary encoder support library
#include <Wire.h>                                        // I2C functions support library
#include <LiquidCrystal.h>                               // LCD non-I2C display functions support library (for Sainsmart LCD/keypad shield or compatible clones)

/***************************************************************************************************
 *                                   User Global Parameters Setup                                  *
 ***************************************************************************************************/

#define ENCODER_TYPE 2                                   // for half-step quardrature encoders like KY-40 set to 2 & full-step encoders set to 1 (you'll know if you need to adjust this!)

#define btn_PORT A0                                      // LCD/keypad shield buttons feed Arduino analog port A0 (resistive voltage divider)

#define btn_RIGHT  0                                     // assign LCD/keypad shield buttons numeric values
#define btn_UP     1
#define btn_DOWN   2
#define btn_LEFT   3
#define btn_SELECT 4
#define btn_NONE   5

#define DETECTOR_MAX_SEL 7                               // SDR# 8 different detector types (AM, CW, DSB, NFM, LSB, USB, WFM, RAW)
#define MENU_MAX_SEL 3                                   // SDR# 4 different menu selections (change frequency, change tuning step, change gain, change detector)

#define VFO_LOWER 24000000                               // RTL-SDR VFO lower frequency limit (adjust as required this is specific to the R820T/T2 tuner)
#define VFO_UPPER 1700000000                             // RTL-SDR VFO upper frequency limit (adjust as required this is specific for the R820T/T2 tuner)

#define GAIN_MIN 25                                      // SDR# minimum gain (25 dB = mute)
#define GAIN_MAX 60                                      // SDR# maximum gain (60 dB)
 
#define DELAY 500                                        // default delay value in milliseconds for human hands & eyes (adjust as required)

#define ROTARY_BUTTON 11                                 // rotary encoder shaft push button switch connected to LCD/keypad shield/Arduiono D11 (change as required)     

int RETVAL;                                              // general purpose return value variable (borrowed from Paradox PAL)
boolean KEYPAD_FLAG = false;                             // keypad flag set if any keypad buttons pressed

Encoder encoder(3, 2);                                   // initialize encoder using digital pin D2 & D3 (switch them if your encoder tunes in the "wrong" direction)

LiquidCrystal lcd(8, 9, 4, 5, 6, 7);                     // map LCD/keypad shield to Arduino pin assignments (shield sits on top of MCU so you don't have to use jumpers)

int lcd_key = 0;                                         // LCD/keypad variable assignments (for keypad button & input return values)
int lcd_keyin = 0;

// create detector type string values used by SDR# (alphabetic order)
int detector_select = 255;                               // detector type selected pointer ">" character

// create frequency tuning step values & LCD string arrays (change as required)
#define MAX_TUNING_STEPS 19                              // maximum number of frequency tuning steps allocated

const long step_values[] = {1, 10, 25, 50, 100, 250, 500, 1000, 2500, 5000, 10000, 25000, 50000, 100000, 250000, 500000, 1000000, 2500000, 5000000, 10000000};
const String step_strings[] = {"1", "10", "25", "50", "100", "250", "500", "1k", "2k5", "5k", "10k", "25k", "50k", "100k", "250k", "500k", "1M", "2M5", "5M", "10M"};

int menu_select = 0;                                     // menu selection (0 = change freuqency, 1 = change tuning step/increment, 2 = change audio gain, 3 = change detector type)
int step_index = 0;                                      // frequency step index pointer
long newPos = 0;                                         // encoder startup current position value
long oldPos = 0;                                         // encoder startup old position value
int power_select = 0;                                    // SDR# RTL-SDR run/power on/off status

//char serial_IO = '-';                                    // status of SDR NetControl connection ('-' not connected, '+' connected)
long frequency;                                          // SDR# frequency
long audio_gain;                                         // SDR# audio gain level      
long tuning_step = 1;                                    // default frequency tuning step/increment (1 Hz)
String status_line;                                      // SDR# status line (tuning step/increment, audio gain and detector type)
String hertz = "1";                                      // default starting VFO frequecy tuning/step increment character string
String detector_type;                                    // SDR# detector type

/***************************************************************************************************
 *                                       Scan LCD/keypad Shield                                    *
 ***************************************************************************************************/
int scan_Keypad() {
  
  lcd_keyin = analogRead(btn_PORT);                      // scan keypad analog port for any input from button/resistor voltage divider

  KEYPAD_FLAG = true;                                    // assume keypad key is/will be pressed (inverted logic)
  
  // these are raw voltage values derived from keypad voltage divider & each keypad may have slightly different values (fairly wide detector window used as a result) 
  // my LCD/keypad shield buttons raw voltage values viewed n serial monitor window: right = 14, left = 527, up = 166, down = 352, select = 756, none = 1023
  if (lcd_keyin > 900) {                                 // no shield buttons pressed
    
    KEYPAD_FLAG = false;                                 // reset keypad flag
    return btn_NONE;                                     // make this 1st option for speed since it's the most likely result (most of the time)
  
  } 
 
  // for v1.1 LCD/keypad shields use these analog values
  if (lcd_keyin < 100) return btn_RIGHT;  
  if (lcd_keyin < 200) return btn_UP; 
  if (lcd_keyin < 400) return btn_DOWN; 
  if (lcd_keyin < 600) return btn_LEFT; 
  if (lcd_keyin < 800) return btn_SELECT;   

} // end scan_Keypad()

/***************************************************************************************************
 *                                       Set SDR# frequency                                        *
 ***************************************************************************************************/
void set_Frequency() {

  if (!KEYPAD_FLAG) {                                    // keypad button press didn't send us here so check rotary encoder instead else drop down to rest of routine
    
    RETVAL = scan_RotaryEncoder();                       // scan rotary encoder for any rotation
  
    if (RETVAL == 0) return;                             // nothing, so return else drop down into rest of routine

  }  

  frequency = frequency + (tuning_step * RETVAL);        // adjust frequency up/down according to current turning step/increment (RETVAL = +1 or -1 so we either add or subtract the tuning step)
      
  if (frequency > VFO_UPPER)                             // can't exceed upper VFO frequenecy limit
    frequency = VFO_LOWER;                               // wrap around to lower VFO frequency limit (implements circular buffer)
  else if (frequency < VFO_LOWER)                        // can't go below lower VFO frequency limit  
    frequency = VFO_UPPER;                               // wrap around to upper VFO frequency limit (implements circular buffer)
   
  // send Net Remote serial command to change frequency on RTL SDR dongle via NetRemote pluging using JSON functions libary
  Serial.print("{\"Command\": \"Set\", \"Method\": \"Frequency\", \"Value\": "); Serial.print(frequency); Serial.println("}");   

  show_Frequency();                                      // update LCD frequency display
  
} // end set_Frequency()

/***************************************************************************************************
 *                                       Get SDR# frequency                                        *
 ***************************************************************************************************/
void get_Frequency() {

  String inData = Serial.readStringUntil('\n');          // flush serial buffer of any stale data (from previous JSON commands) & must be done each & every time!
  
  // get SDR#'s tuned frequency using JSON function via serial connection
  Serial.println("{\"Command\": \"Get\", \"Method\": \"Frequency\"}");  
    inData = Serial.readStringUntil('\n');               // get answer back
  
  StaticJsonBuffer<200> jsonBuffer;                      // use Arduino JSON fuction libary to parse for result
    JsonObject& root = jsonBuffer.parseObject(inData);

  frequency = root["Value"];                             // if we aren't connected/sync with RTL-SDR dongle frequency = 0 Hz (press RST/reset button to reconnect via NetRemote plugin
  
//  if (frequency <= 0)                                    // no valid frequency so we can't be connected to RTL-SDR dongle via SDR NetControl
//    serial_IO = '-';a
//  else 
//    serial_IO = '+';
      
  if (frequency > VFO_UPPER)                             // SDR# can tune RTL-SDR beyond our RTL-SDR's tuning limits so adjust as required (also serial I/O errors can cause bad data)
    frequency = VFO_UPPER;
  else if (frequency < VFO_LOWER)
    frequency = VFO_LOWER;  
  
  set_Frequency();                                       // set SDR# to legal RTL-SDR (VFO) frequency 
 
  encoder.write(0);                                      // reset encoder counter to known starting point

} // end get_Frequency()

/***************************************************************************************************
 *                                  Show Frequency on LCD Display                                  *
 ***************************************************************************************************/
void show_Frequency(){

  int millions = int(frequency / 1000000);               // breakup receiver frequency into individual components
  int hundredthousands = (frequency / 100000) % 10;
  int tenthousands = (frequency / 10000) % 10;
  int thousands = (frequency / 1000) % 10;
  int hundreds = (frequency / 100) % 10;
  int tens = (frequency / 10) % 10;
  int ones = (frequency / 1) % 10;
                                                        
  //lcd.setCursor(0, 0);// lcd.print("                ");    // home LCD cursor & clear frequency display 
    lcd.setCursor(1, 0);                                 // resposition cursor & display each frequency byte (individual digit)    

  //lcd.print(serial_IO);                                  // display status of SDR NetControl connnection
  
  if (millions > 999) {                                  // gigahertz frequency?
    millions = millions - 1000;                          // yes, adjust millions value & pad display with leading 1 gigahertz digit
    lcd.print("001.");
  }
  else
    lcd.print("000.");                                   // no, pad display with leading 0's

  if (millions < 10)                                     // do we need to add extra 0's for padding?
    lcd.print("00");
  else if (millions < 100)
    lcd.print("0");
         
  lcd.print(millions);                                   // display frequency individual components on LCD
    lcd.print(".");
  
  lcd.print(hundredthousands); lcd.print(tenthousands); lcd.print(thousands); lcd.print(".");
    lcd.print(hundreds); lcd.print(tens); lcd.print(ones);
  
  //show_Status();                                         // show/update LCD status line
  
} // end show_Frequency()

/***************************************************************************************************
 *                                       Get SDR# audio gain                                       *
 ***************************************************************************************************/
void get_AudioGain() {

  String inData = Serial.readStringUntil('\n');          // flush serial input buffer of any stale data from previous JSON commands
  
  // get SDR#'s volume setting 25 dB (mute) to 60 dB using JSON function via serial port connection
  Serial.println("{\"Command\": \"Get\", \"Method\": \"AudioGain\"}");
    inData = Serial.readStringUntil('\n');               // get answer back
   
  StaticJsonBuffer<200> jsonBuffer;                      // use Arduino JSON fuction libary to parse for result
    JsonObject& root = jsonBuffer.parseObject(inData);

  audio_gain = root["Value"];                            // if we aren't connected/sync with RTL-SDR dongle frequency = 0 Hz (press RST/reset button to reconnect via NetRemote plugin

  if (audio_gain <= 0)                                   // not connected to SDR# so set gain to minimum level
    audio_gain = GAIN_MIN;
    
} // end get_AudioGain()

/***************************************************************************************************
 *                                        Set SDR# audio gain                                      *
 ***************************************************************************************************/
void set_AudioGain() {

  if (!KEYPAD_FLAG) {                                    // keypad button press didn't send us here so check rotary encoder instead else drop down to rest of routine
    
    RETVAL = scan_RotaryEncoder();                       // scan rotary encoder for any rotation
  
    if (RETVAL == 0) return;                             // nothing, so return else drop down into rest of routine

  }  

  audio_gain += RETVAL;                                  // add RETVAL value (+1 or -1) either rotary encoder or keypad has set RETVAL value (+1 or -1)
  
  if (audio_gain > GAIN_MAX)                             // can't exceed wrap around to min value (circular buffer)
    audio_gain = GAIN_MIN;                             
  else if (audio_gain < GAIN_MIN)                        // can't go below so wrap around to max value (circular buffer)
    audio_gain = GAIN_MAX;                    

  // set SDR# audio gain level (in dB)
  Serial.print("{\"Command\": \"Set\", \"Method\": \"AudioGain\", \"Value\": "); Serial.print(audio_gain); Serial.println("}");   

  show_Status();                                         // show/update LCD status line

} // end set_AudioGain()

/***************************************************************************************************
 *                                       Get SDR# Detector Type                                    *
 ***************************************************************************************************/
void get_DetectorType() {

  String inData = Serial.readStringUntil('\n');          // flush serial input buffer of any stale data from previous JSON commands
  
  // request current SDR# detector type
  Serial.println("{\"Command\": \"Get\", \"Method\": \"DetectorType\"}");
    inData = Serial.readStringUntil('\n');               // get answer back
   
  StaticJsonBuffer<200> jsonBuffer;                      // use Arduino JSON fuction libary to parse for result
    JsonObject& root = jsonBuffer.parseObject(inData);
  
  detector_select = 255;                                 // reset detector_select integer value
  
  const char * DETECTOR_TYPE = root["Value"];            // get JSON detector type const char pointer value
    detector_type = String(DETECTOR_TYPE);               // convert to const char pointer to string value
  
  if (detector_type == "NFM")                            // get detector type string value and set select pointer index
    detector_select = 0;
  else if (detector_type == "AM") 
    detector_select = 1;
  else if (detector_type == "LSB")
    detector_select = 2;
  else if (detector_type == "USB")
    detector_select = 3;
  else if (detector_type == "WFM")
    detector_select = 4;
  else if (detector_type == "DSB")
    detector_select = 5;
  else if (detector_type == "CW")
    detector_select = 6;
  else if (detector_type == "RAW")
    detector_select = 7;
  else  
    detector_type == "   ";
    
} // end get_DetectorType()


/***************************************************************************************************
 *                                       Set SDR# Detector Type                                    *
 ***************************************************************************************************/
void set_DetectorType() {

  if (!KEYPAD_FLAG) {                                    // keypad button press didn't send us here so check rotary encoder instead else drop down to rest of routine
    
    RETVAL = scan_RotaryEncoder();                       // scan rotary encoder for any rotation
  
    if (RETVAL == 0) return;                             // nothing, so return else drop down into rest of routine

  }  

  detector_select += RETVAL;                             // add RETVAL value (+1 or -1) either rotary encoder or keypad has set RETVAL value (+1 or -1)
      
  if (detector_select < 0)                               // keep detector type select point in proper range (implements circular buffer)
    detector_select = DETECTOR_MAX_SEL;
  else if (detector_select > DETECTOR_MAX_SEL)
    detector_select = 0;
  
  switch (detector_select) {                             // determine detector_type from detector_select index pointer
    case 0: 
      detector_type = "NFM"; 
      break;
    case 1:
      detector_type = "AM";
      break;
    case 2:
      detector_type = "LSB";
      break;
    case 3:
      detector_type = "USB";
      break;
    case 4:
      detector_type = "WFM";
      break;
    case 5:
      detector_type = "DSB";
      break;
    case 6:
      detector_type = "CW";
      break;
    case 7:
      detector_type = "RAW";
      break; 
    default:
      detector_type = "   ";
      break;

  }

  show_Status();                                         // show/update LCD status line
    
  // set SDR# detector type note special formating required for JSON const char pointer values from string value detector_type
  Serial.print("{\"Command\": \"Set\", \"Method\": \"DetectorType\", \"Value\": "); Serial.print("\"" + detector_type + "\""); Serial.println("}");   

} // end set_DetectorType()

/***************************************************************************************************
 *                                        Show SDR# Status Line                                    *
 ***************************************************************************************************/
void show_Status() {

  lcd.setCursor(0, 0);                                   // reposition LCD cursor & clear out old ">" & status line values
    lcd.print(" ");             
  lcd.setCursor(0, 1);                                   
    lcd.print(" T      G       ");

  switch (menu_select) {                                 // indicate which selection is active & reposition ">" pointer
    case 0:
      lcd.setCursor(0, 0); lcd.print(">");               // point to frequency
      break;
    case 1:
      lcd.setCursor(0, 1); lcd.print(">");               // point to tuning step
      break;  
    case 2:
      lcd.setCursor(7, 1); lcd.print(">");               // point to audio gain
      break;
    case 3:
      lcd.setCursor(12, 1); lcd.print(">");              // point to detector type
      break;
 
  }

  lcd.setCursor(2, 1);                                   // show tuning step/increment value
    lcd.print(hertz); 

  lcd.setCursor(9, 1);
    lcd.print(audio_gain);
  
  lcd.setCursor(13, 1);
    lcd.print(detector_type);

}

/***************************************************************************************************
 *                    Set Frequency Tuning Step/Increment & Show on LCD Display                    *
 ***************************************************************************************************/
void set_TuningStep(){
  
  if (!KEYPAD_FLAG) {                                    // keypad button press didn't send us here so check rotary encoder instead else drop down to rest of routine
    
    RETVAL = scan_RotaryEncoder();                       // scan rotary encoder for any rotation
  
    if (RETVAL == 0) return;                             // nothing, so return else drop down into rest of routine

  }  

  step_index += RETVAL;                                  // add RETVAL value (+1 or -1) either rotary encoder or keypad has set RETVAL value (+1 or -1)

  if (step_index > MAX_TUNING_STEPS)                     // check for upper [frequency] & lower step ranges & adjust using circular buffer
    step_index = 0;                                      // set to first tuning step
  else if(step_index < 0)
    step_index = MAX_TUNING_STEPS;                       // set to last tuning step
  
  tuning_step = step_values[step_index];                 // get step increment to use

  hertz = step_strings[step_index];                      // get frequency step string to display on LCD  

  show_Status();                                         // update SDR# status line

}  // end set_TuningStep()


/***************************************************************************************************
 *                                      Scan Rotary Encoder                                        *
 ***************************************************************************************************/
int scan_RotaryEncoder() {

  RETVAL = 0;                                            // assume no rotation (inverted logic)
  
  newPos = encoder.read() / ENCODER_TYPE;                // scan rotary encoder for any rotation & correct for encoder type (divide by 2 for half-step encoders)
  
  if (newPos != oldPos) {                                // did we rotate encoder?

    if (oldPos > newPos)                                 // rotated left?
      RETVAL = -1;                                       // set RETVAL
    
    else if (oldPos < newPos)                            // rotated right?
      RETVAL = 1;                                        // set RETVAL
   
    oldPos = newPos;                                     // save encoder's new position

  }
  
  return RETVAL;

} // end scan_RotaryEncoder()

/***************************************************************************************************
 *                                 Get Keypad/Encoder Button Press                                 *
 ***************************************************************************************************/
void get_Buttons() {
  
  if (digitalRead(ROTARY_BUTTON) == LOW) {               // was rotary encoder shaft button pressed?
    
    delay(DELAY);                                        // wait to see if long shaft button press for SDR# toggle start/stop function (SDR# must be loaded)
    
    if (digitalRead(ROTARY_BUTTON) == LOW) {             // still pressing encoder shaft button down?
  
      String inData = Serial.readStringUntil('\n');      // flush serial input buffer of any stale data from previous JSON commands
  
      // is SDR# active (playing)?
      Serial.println("{\"Command\": \"Get\", \"Method\": \"IsPlaying\"}");
        inData = Serial.readStringUntil('\n');           // get answer back
  
      StaticJsonBuffer<200> jsonBuffer;                  // use Arduino JSON fuction libary to parse for result
        JsonObject& root = jsonBuffer.parseObject(inData);
  
      boolean STATUS = root["Value"];                    // get JSON detector type const char pointer value (0 = not playing, 1 = yes, is playing)

      if (STATUS)                                        // if SDR# is playing stop it else start it (SDR# must be loaded)
        Serial.println("{\"Command\": \"Exe\", \"Method\": \"Stop\"}");        
      else
        Serial.println("{\"Command\": \"Exe\", \"Method\": \"Start\"}");

      //digitalWrite(ROTARY_BUTTON, HIGH);               // clear rotary button press by reseting to HIGH
 
      delay(DELAY);                                      // add extra long delay so we don't reset what we just did!
        delay(DELAY);
        
    }
    
    else {                                               // pressing rotary encoder shaft scrolls through functions (change frequency, change tuning step, change gain, change detector)
       
       menu_select++;                                    // next menu selection
       
       if (menu_select > MENU_MAX_SEL)                   // implements circular buffer to move circuitously through all 4 options
          menu_select = 0;                      

        show_Status();                                   // display LCD status line
    }
  
  }

  if ((lcd_key = scan_Keypad()) != btn_NONE) {           // see if any LCD/keypad shield buttons pressed
    
    switch (lcd_key) {                                   // determine what to do with any keypad keypress value
 
      case btn_RIGHT:                                    // to do
        break;
     
      case btn_LEFT:                                     // to do
        break;

      case btn_UP:

        delay(DELAY);
        
        RETVAL = 1;                                      // tuning step up
      
        break;
       
      case btn_DOWN:
        
        delay(DELAY);
        
        RETVAL = -1;                                     // tuning step down
     
        break;
        
      case btn_SELECT:                                   // menu selection (tuning step/increment, audio gain, detector type)

        delay(DELAY);                                    // debounce delay

        menu_select++;
        
        if (menu_select > MENU_MAX_SEL)                  // pressing rotary encoder shaft scrolls through functions (change frequency, change tuning step, change gain, change detector)
          menu_select = 0;                               // implements circular buffer to move circuitously through all 4 options

        show_Status();
        
        break;
     
    } // end select...case

  } // end if
      
} // end get_Buttons()

/***************************************************************************************************
 *                                  Setup Program Runtime Requirements                             *
 ***************************************************************************************************/
void setup() {

  Serial.begin(115000);                                  // default serial I/O speed for SDR# NetRemote plugin
  
  pinMode(ROTARY_BUTTON, INPUT);                         // rotary encoder shaft push button connected to Ardunio Uno analog input pin A3 used for input
    digitalWrite(ROTARY_BUTTON, HIGH);                   // set status to high (open or not pushed/pressed)
 
  lcd.begin(16, 2);                                      // initialize LCD display (16 characters by 2 lines) & clear & home cursor
  
  get_Frequency();                                       // get SDR#'s current tuned frequency (press RST reset button to resync with SDR#)
 
  get_AudioGain();                                       // get SDR#'s current volume control setting
 
  get_DetectorType();                                    // get SDR#'s current demodulator type (AM/CW/DSB/LSB/USB/NFM/RAW/WFM)
  
  show_Frequency();

  show_Status();                                         // show current SDR# settings (if connected) else defaults
  
} // end setup()


/***************************************************************************************************
 *                                      Main Program COntrol Loop                                  *
 ***************************************************************************************************/
void loop() {

  switch (menu_select) {                                 // detemine what to do based om menu selection
    
    case 0:
        set_Frequency();
        break;
    case 1:
        set_TuningStep();
        break;
    case 2:
        set_AudioGain();
        break;
    case 3:
        set_DetectorType();
        break;                                               
  
  }

  get_Buttons();                                         // check for any rotary encoder or LCD keypad button presses
  
} // end loop
